def find_pipeline_digest(pipelines, digest):
    if type(digest) != list:
        digest = [digest]
    digestMap = {d: True for d in digest}
    ret = []
    for pipeline in pipelines:
        if pipeline['pipeline_digest'] in digestMap:
            ret.append(pipeline)
    return ret


import pkg_resources
import string
import numpy as np
from dateutil.parser import parse
import json
import networkx as nx
from ._graph_matching import pipeline_to_graph, merge_multiple_graphs
from ._powerset_analysis import compute_group_importance
from ._comm_api import setup_comm_api
from collections import defaultdict
import copy

exportedPipelines = []

def comm_powerset_analysis(msg):
    pipelines = msg['pipelines']
    scores = msg['scores']
    analysis = compute_group_importance(pipelines, scores, 2)
    return {"analysis": analysis}
setup_comm_api('powerset_analysis_comm_api', comm_powerset_analysis)

def comm_merge_graphs(msg):
    pipelines = msg['pipelines']
    graphs = [pipeline_to_graph(pipeline, pipeline['pipeline_digest']) for pipeline in pipelines]
    merged = merge_multiple_graphs(graphs)
    data_dict = nx.readwrite.json_graph.node_link_data(merged)
    return {"merged": data_dict}
setup_comm_api('merge_graphs_comm_api', comm_merge_graphs)

def comm_export_pipelines(msg):
    global exportedPipelines
    exportedPipelines = msg['pipelines']
    return {}
setup_comm_api('export_pipelines_comm_api', comm_export_pipelines)

def get_exported_pipelines():
    global exportedPipelines
    return exportedPipelines


def id_generator(size=15):
    """Helper function to generate random div ids. This is useful for embedding
    HTML into ipython notebooks."""
    chars = list(string.ascii_uppercase)
    return ''.join(np.random.choice(chars, size, replace=True))


def make_html(data_dict, id):
	lib_path = pkg_resources.resource_filename(__name__, "build/pipelineVis.js")
	bundle = open(lib_path, "r", encoding="utf8").read()
	html_all = """
	<html>
	<head>
	</head>
	<body>
	    <script>
	    {bundle}
	    </script>
	    <div id="{id}">
	    </div>
	    <script>
	        pipelineVis.renderPipelineMatrixBundle("#{id}", {data_dict});
	    </script>
	</body>
	</html>
	""".format(bundle=bundle, id=id, data_dict=json.dumps(data_dict))
	return html_all

def extract_primitive_names(pipeline):
    return [s['primitive']['python_path'] for s in pipeline['steps']]

def extract_module_matrix(pipelines):
    from sklearn import preprocessing
    import numpy as np
    le = preprocessing.LabelEncoder()
    all_primitives = set()
    for pipeline in pipelines:
        all_primitives = all_primitives.union(extract_primitive_names(pipeline))
    le.fit(list(all_primitives))
    data_matrix = np.zeros([len(pipelines), len(le.classes_)])
    for i, pipeline in enumerate(pipelines):
        for j, primitive in enumerate(extract_primitive_names(pipeline)):
            idx_primitive = le.transform([primitive])
            data_matrix[i, idx_primitive] = 1
    return data_matrix, le.classes_

def extract_scores(pipelines):
    scores = []
    for pipeline in pipelines:
        score = pipeline['scores'][0]['value']
        scores.append(score)
    return np.array(scores)

def rename_pipelines(pipelines):
    sourceMap = defaultdict(lambda: 1)
    for pipeline in pipelines:
        source = pipeline['pipeline_source']['name']
        pipeline['pipeline_source']['name'] = '{} #{}'.format(source, sourceMap[source])
        sourceMap[source] += 1


def transform_module_type(module_type):
    map = {
        'feature_extraction': 'Feature Extraction',
        'learner': 'Classification',
        'normalization': 'Preprocessing',
        'feature_construction': 'Feature Selection',
        'classification': 'Classification',
        'data_transformation': 'Preprocessing',
        'schema_discovery': 'Preprocessing',
        'data_preprocessing': 'Preprocessing',
        'data_cleaning': 'Preprocessing',
        'regression': 'Regression',
        'operator': 'Operator',
        'feature_selection': 'Feature Selection',
        'semisupervised_classification': 'Classification',
        'natural_language_processing': 'NLP',
        'time_series_forecasting': 'Forecasting',
        'time_series_classification': 'TS Classification',
        'time_series_segmentation': 'Feature Extraction'
    }
    if module_type in map:
        return map[module_type]
    else:
        return ' '.join([n.capitalize() for n in module_type.split("_")])

def extract_primitive_info(pipelines, enet_alpha, enet_l1):
    module_matrix, module_names = extract_module_matrix(pipelines)
    scores = extract_scores(pipelines)
    infos = {}
    module_types = set()
    for pipeline in pipelines:
        for step in pipeline['steps']:
            python_path = step['primitive']['python_path']
            if python_path in infos:
                continue
            split = python_path.split(".")
            module_desc = step['primitive']['name']
            module_type = transform_module_type(split[2])
            module_types.add(module_type)
            module_name = split[3]
            infos[python_path] = {
                "module_desc": module_desc,
                "module_type": module_type,
                "module_name": module_name,
            }
    return infos, module_types

def extract_d3m_time_metric(pipelines):
    for pipeline in pipelines:
        if 'end' in pipeline and 'start' in pipeline: # using D3M prediction time format
            diff = parse(pipeline['end']) - parse(pipeline['start'])
            diff_sec = diff.total_seconds()
            pipeline['scores'].append({
                'metric': {'metric': 'PRED TIME (s)'},
                'normalized': diff_sec,
                'value': diff_sec,
            })

def compute_metric_map(pipelines):
    for pipeline in pipelines:
        pipeline['score_map'] = {}
        for score in pipeline['scores']:
            pipeline['score_map'][score['metric']['metric']] = score            

def prepare_data_pipeline_matrix(pipelines, enet_alpha=0.001, enet_l1=0.1):
    pipelines = copy.deepcopy(pipelines)
    extract_d3m_time_metric(pipelines)
    compute_metric_map(pipelines)
    pipelines = sorted(pipelines, key=lambda x: x['scores'][0]['normalized'], reverse=True)
    rename_pipelines(pipelines)
    info, module_types = extract_primitive_info(pipelines, enet_alpha=enet_alpha, enet_l1=enet_l1)
    data = {
        "infos": info,
        "pipelines": pipelines,
        "module_types": list(module_types),
    }
    return data


def plot_pipeline_matrix(pipelines):
    from IPython.core.display import display, HTML
    id = id_generator()
    data_dict = prepare_data_pipeline_matrix(pipelines)
    html_all = make_html(data_dict, id)
    display(HTML(html_all))



#!/usr/bin/env python
# coding: utf-8

# Script that preprocesses and merges "pipelines" and "pipeline_runs" from D3M Metalearning 
# Last files downloaded from https://metalearning.datadrivendiscovery.org/dumps/2019/10/04/
# Metalearning data description: https://gitlab.com/datadrivendiscovery/metalearning

import json
import argparse

def merge_pipeline_files(pipelines_file, pipeline_runs_file, n=-1, verbose=False):
    # Function that merges the pipelines file with the pipeline_runs file.
    # Arguments:
    # pipelines_file: Path to the pipeline_runs file. See http://metalearning.datadrivendiscovery.org/dumps
    # pipeline_runs_file: Path to the pipelines file. See http://metalearning.datadrivendiscovery.org/dumps
    # n: Number of merged pipelines to output. If n=-1, save all pipelines to the merged file

    # adding pipelines to lookup table {<digest>: <data>}
    print ("Adding pipelines to lookup table...")
    pipelines = {}
    with open(pipelines_file, "r", encoding="utf8") as f:
        for line in f:
            pipeline = json.loads(line)
            pipelines[pipeline["digest"]] = pipeline
    
    # merging pipeline information with pipeline_runs
    print ("Merging pipeline information with pipeline_runs_file (this might take a while)...")
    merged = []
    with open(pipeline_runs_file, "r",  encoding="utf8") as f:
        for line in f:
            if len(merged) == n:
                break
            try:
                run = json.loads(line)
                if run['run']['phase']!='PRODUCE':
                    continue
                pipeline = pipelines[run["pipeline"]["digest"]]
                data = {
                    'pipeline_id': pipeline['id'],
                    'pipeline_digest': pipeline['digest'],
                    'pipeline_source': pipeline['source'],
                    'created': pipeline['created'],
                    'inputs': pipeline['inputs'],
                    'outputs': pipeline['outputs'],
                    'problem': run['problem'],
                    'start': run['start'],
                    'end': run['end'],
                    'steps': pipeline['steps'],
                    'scores': run['run']['results']['scores']
                }
                merged.append(data)
            except Exception as e:
                if (verbose):
                    print (repr(e))
        print ("Done.")
    return merged



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("pipeline_runs_file", help="Path to the pipeline_runs file. See http://metalearning.datadrivendiscovery.org/dumps", type=str)
    parser.add_argument("pipelines_file", help="Path to the pipelines file. See http://metalearning.datadrivendiscovery.org/dumps", type=str)
    parser.add_argument("output_file", help="Path to output file.", type=str)
    parser.add_argument("-n", "--number_pipelines", help="Number of pipelines to save to the file. If n=-1, save all pipelines to the merged file.", type=int, default=-1)
    parser.add_argument("-v", "--verbose", help="Increase output verbosity (show json key errors)", action="store_true")
    args = parser.parse_args()
    d = merge_pipeline_files(args.pipelines_file, args.pipeline_runs_file, n = args.number_pipelines, verbose=args.verbose)
    with open(args.output_file, "w", encoding="utf8") as f:
        json.dump(d, f)


import networkx as nx
import numpy as np
from collections import defaultdict

def node_type(g, node):
    node = g.nodes[node]
    path_split = node['data'][0]['python_path'].split('.')
    if (len(path_split) > 2):
        return path_split[2]
    else:
        return node['data'][0]['python_path']

def compute_edge_type(g, edge):
    return '.'.join([node_type(g, edge[0]), node_type(g, edge[1])])

def create_pairwise_conn_graph(g1, g2):
    PCG = nx.DiGraph()
    for edge_g1 in g1.edges:
        for edge_g2 in g2.edges:
            type_edge_g1 = compute_edge_type(g1, edge_g1)
            type_edge_g2 = compute_edge_type(g2, edge_g2)
            if (type_edge_g1 == type_edge_g2): # edge the same type
                s1, d1 = edge_g1
                s2, d2 = edge_g2
                PCG.add_edge((s1, s2), (d1, d2), edge_type = type_edge_g1)
                PCG.add_edge((d1, d2), (s1, s2), edge_type = type_edge_g1)
    return PCG

def transform_induced_propagation_graph(pcg):
    edges = pcg.edges(data=True)
    edge_dict = defaultdict(lambda : defaultdict(lambda: []))
    # creating edge_dict information
    for edge in edges:
        n1 = edge[0]
        n2 = edge[1]
        data = edge[2]
        edge_type=data['edge_type']
        edge_dict[n1][edge_type].append(edge)
    
    edge_weight_map = {}
    
    for node in pcg.nodes:
        for edge_type in edge_dict[node]:
            n_edges_same_type = len(edge_dict[node][edge_type])
            weight = 1/n_edges_same_type
            for edge in edge_dict[node][edge_type]:
                edge_weight_map[edge[:2]] = weight
    nx.set_edge_attributes(pcg, edge_weight_map, 'weight')
    return pcg

def similarity_flooding(similarity_matrix, g1, g2, alpha = 0.01, n_iter = 100):
    pcg = create_pairwise_conn_graph(g1, g2)
    ipg = transform_induced_propagation_graph(pcg)
    g1_node_map = {node:idx for idx, node in enumerate(g1.nodes)}
    g2_node_map = {node:idx for idx, node in enumerate(g2.nodes)}
    adj = nx.adjacency_matrix(pcg, weight='weight').todense()
    
    #setting sim weights to ipg graph
    for node in ipg.nodes:
        g1_idx = g1_node_map[node[0]]
        g2_idx = g2_node_map[node[1]]
        ipg.nodes[node]['similarity'] = similarity_matrix[g1_idx, g2_idx]
    
    #creating similarity vector
    similarity_vector = np.reshape(np.array([node[1]['similarity'] for node in ipg.nodes(data=True)]), [-1,1])
    similarity_vector = similarity_vector / np.max(similarity_vector)
    for it in range(n_iter):
        similarity_vector = alpha * np.dot(adj, similarity_vector) + (1-alpha) * similarity_vector
        similarity_vector = similarity_vector / np.max(similarity_vector)
    
    #retrieving sim weights to similarity_matrix
    for node in ipg.nodes:
        g1_idx = g1_node_map[node[0]]
        g2_idx = g2_node_map[node[1]]
        similarity_matrix[g1_idx, g2_idx] = ipg.nodes[node]['similarity']
    
    return similarity_matrix




def setup_comm_colab(api_call_id, callback):
    # Function that connects javascript call with Colab Notebook
    from google.colab import output
    from IPython import display
    def _recv(msg):
        return display.JSON(callback(msg)) # Use display.JSON to transfer an object
    output.register_callback(api_call_id, _recv)

def setup_comm_jupyter(api_call_id, callback): 
    # Function that connects javascript call with Jupyter Notebook
    def _comm_api(comm, open_msg): 
        @comm.on_msg
        def _recv(msg):
            ret = callback(msg['content']['data'])
            comm.send(ret)
    get_ipython().kernel.comm_manager.register_target(api_call_id, _comm_api)

def setup_comm_api(api_call_id, callback):
    # Function that abstracts notebook connection to javascript
    try:
        jupyter_setup = True
        setup_comm_jupyter(api_call_id, callback)
    except Exception:
        jupyter_setup = False
    try:
        colab_setup = True
        setup_comm_colab(api_call_id, callback)
    except Exception:
        colab_setup = False
    if not jupyter_setup and not colab_setup:
        print("Error: Cannot find Jupyter/Colab namespace for Python")


from itertools import combinations, chain
from collections import defaultdict
from sklearn import preprocessing
import numpy as np
from scipy.stats import pointbiserialr

def extract_primitive_names(pipeline):
    return [s['primitive']['python_path'] for s in pipeline['steps']]

def extract_primitive_matrix(pipelines):
    le = preprocessing.LabelEncoder()
    all_primitives = set()
    for pipeline in pipelines:
        all_primitives = all_primitives.union(extract_primitive_names(pipeline))
    le.fit(list(all_primitives))
    data_matrix = np.zeros([len(pipelines), len(le.classes_)], dtype=int)
    for i, pipeline in enumerate(pipelines):
        for j, primitive in enumerate(extract_primitive_names(pipeline)):
            idx_primitive = le.transform([primitive])
            data_matrix[i, idx_primitive] = 1
    return data_matrix, le.classes_

def extract_scores(pipelines):
    scores = []
    for pipeline in pipelines:
        score = pipeline['scores'][0]['value']
        scores.append(score)
    return scores


def compute_group_importance(pipelines, scores, up_to_k=5):
    primitive_matrix, primitives = extract_primitive_matrix(pipelines)
    column_idx_map = {p: idx for idx, p  in enumerate(primitives)}
    importances = {}
    for k in range(1, up_to_k+1):
        for selected_columns in combinations(primitives, k):
            selected_columns_idx = np.array([column_idx_map[c] for c in selected_columns])
            sub_matrix = primitive_matrix[:, selected_columns_idx]
            used_all = np.prod(sub_matrix, axis = 1)
            importance, _  = pointbiserialr(used_all, scores)
            importance = 0 if np.isnan(importance) else importance
            importances[frozenset(selected_columns)] = importance

    # keeping only the ones for which importance is greater than those of its component parts
    kept = []
    for selected_columns in importances.keys():
        if (len(selected_columns) > 1):
            to_add = True
            for subgroup in chain(*[list(combinations(selected_columns, take)) for take in range(1, len(selected_columns))]):
                if abs(importances[frozenset(subgroup)]) >= abs(importances[selected_columns]):
                    to_add = False
                    break
            if to_add:
                kept.append({'importance':importances[selected_columns], 'group': list(selected_columns)})
    return sorted(kept, key=lambda x:abs(x['importance']), reverse=True)


import numpy as np

class DefaultOrder:
    def __init__(self, order):
        self.order_map = {}
        for idx, elem in enumerate(order):
            self.order_map[elem] = idx
    def __call__(self, elem):
        if elem not in self.order_map:
            self.order_map[elem] = len(self.order_map)
        return self.order_map[elem]

def find_metric_name(automl):
    try:
        statistics = automl.sprint_statistics()
        lines = statistics.split("\n")
        my_line = list(filter(lambda x: x.upper().find("METRIC") > -1, lines))[0]
        metric = my_line.split(":")[1].strip().upper()
        return metric
    except Exception as e:
        return "METRIC"

def find_ensemble_weights(all_params, models_with_weights):
    weights = np.zeros(len(all_params))
    try:
      for idx, param in enumerate(all_params):
          for weight, pipeline in models_with_weights:
              weighted_param = pipeline.config.get_dictionary()
              if param == weighted_param:
                  weights[idx] = weight
                  break
    except Exception:
        pass
    return weights

def import_autosklearn(automl, source='auto-sklearn'):
    cv_results = automl.cv_results_
    weights = find_ensemble_weights(automl.cv_results_['params'], automl.get_models_with_weights())
    node_order = DefaultOrder(['balancing', 'data_preprocessing', 'feature_preprocessor','classifier', 'regressor'])
    n_models = len(cv_results['mean_test_score'])
    pipelines = []
    metric_name = find_metric_name(automl)
    for i in range(n_models):
        # Computing auxiliary pipeline structure
        struct = {}
        for key in cv_results['params'][i]:
            split = key.split(":")
            if split[0] == 'balancing' and split[1] == 'strategy':
                if cv_results['params'][i][key] != 'none':
                    struct['balancing'] = {'class_balancing': {'strategy': cv_results['params'][i][key]}}
                continue
            if split[1] == '__choice__':
                continue
            module_type = split[0]
            if module_type not in struct:
                struct[module_type] = {}    
            module_name = split[1]
            if module_name not in struct[module_type]:
                struct[module_type][module_name] = {}
            module_param = '_'.join(filter(lambda x: x != '__choice__', split[2:]))
            struct[module_type][module_name][module_param] = cv_results['params'][i][key]
        ordered_types = struct.keys()
        ordered_types = sorted(ordered_types, key=lambda x: node_order(x))

        # Creating pipeline graph
        pipeline = {
            'inputs': [{'name': 'input dataset'}],
            'steps': [],
            'scores': [{
                'metric': {'metric': metric_name, 'params': {'pos_label': '1'}},
                'normalized': cv_results['mean_test_score'][i],
                'value': cv_results['mean_test_score'][i],
            },
            {
                'metric': {'metric': 'ENSEMBLE WEIGHT', 'params': {'pos_label': '2'}},
                'normalized': weights[i],
                'value': weights[i]
            },{
                'metric': {'metric': 'MEAN FIT TIME', 'params': {'pos_label': '3'}},
                'normalized': cv_results['mean_fit_time'][i],
                'value': cv_results['mean_fit_time'][i],
            }],
            'pipeline_source': {'name': source},
            'pipeline_digest': '{}'.format(i),
        }
        prev_list = ['inputs.0']
        cur_step_idx = 0
        for module_type in ordered_types:
            new_prev_list = []
            modules = struct[module_type]
            for module in modules:
                step_ref = 'steps.{}.produce'.format(cur_step_idx)
                cur_step_idx += 1
                new_prev_list.append(step_ref)
                step = {
                    'primitive': {'python_path':'auto_sklearn.primitives.{}.{}'.format(module_type, module), 'name': module},
                    'arguments':{},
                    'outputs': [{'id': 'produce'}],
                    'reference': {'type': 'CONTAINER', 'data': step_ref},
                    'hyperparams': {}
                }
                for param in struct[module_type][module]:
                    step['hyperparams'][param] = {
                        'type': 'VALUE',
                        'data': struct[module_type][module][param]
                    }                
                for idx, prev in enumerate(prev_list):
                    cur_argument_idx = 'input{}'.format(idx)
                    step['arguments'][cur_argument_idx] = {
                        'data': prev
                    }
                pipeline['steps'].append(step)
            prev_list = new_prev_list
        pipeline['outputs'] = []
        for prev in prev_list:
            pipeline['outputs'].append({'data': prev})        
        pipelines.append(pipeline)
    return pipelines

def import_autosklearn_test_data(automl, test_X, test_y, metric_eval, source='auto-sklearn'):
    node_order = DefaultOrder(['balancing', 'data_preprocessing', 'feature_preprocessor','classifier', 'regressor'])
    models_with_weights = automl.get_models_with_weights()
    weights = []
    models = []
    for weight, model in models_with_weights:
        weights.append(weight)
        models.append(model)
    n_models = len(models)
    pipelines = []
    for i in range(n_models):
        # Computing auxiliary pipeline structure
        config = models[i].config.get_dictionary()
        struct = {}
        for key in config:
            split = key.split(":")
            if split[0] == 'balancing' and split[1] == 'strategy':
                if config[key] != 'none':
                    struct['balancing'] = {'class_balancing': {'strategy': config[key]}}
                continue
            if split[1] == '__choice__':
                continue
            module_type = split[0]
            if module_type not in struct:
                struct[module_type] = {}
            module_name = split[1]
            if module_name not in struct[module_type]:
                struct[module_type][module_name] = {}
            module_param = '_'.join(filter(lambda x: x != '__choice__', split[2:]))
            struct[module_type][module_name][module_param] = config[key]
        ordered_types = struct.keys()
        ordered_types = sorted(ordered_types, key=lambda x: node_order(x))

        # Computing score
        pred_y = models[i].predict(test_X)
        score = metric_eval(test_y, pred_y)

        # Creating pipeline graph
        pipeline = {
            'inputs': [{'name': 'input dataset'}],
            'steps': [],
            'scores': [{
                'metric': {'metric': str(metric_eval).capitalize(), 'params': {'pos_label': '1'}},
                'normalized': score,
                'value': score,
            },
            {
                'metric': {'metric': 'Ensemble Weight', 'params': {'pos_label': '2'}},
                'normalized': weights[i],
                'value': weights[i]
            }],
            'pipeline_source': {'name': source},
            'pipeline_digest': '{}'.format(i),
        }
        prev_list = ['inputs.0']
        cur_step_idx = 0
        for module_type in ordered_types:
            new_prev_list = []
            modules = struct[module_type]
            for module in modules:
                step_ref = 'steps.{}.produce'.format(cur_step_idx)
                cur_step_idx += 1
                new_prev_list.append(step_ref)
                step = {
                    'primitive': {'python_path':'auto_sklearn.primitives.{}.{}'.format(module_type, module), 'name': module},
                    'arguments':{},
                    'outputs': [{'id': 'produce'}],
                    'reference': {'type': 'CONTAINER', 'data': step_ref},
                    'hyperparams': {}
                }
                for param in struct[module_type][module]:
                    step['hyperparams'][param] = {
                        'type': 'VALUE',
                        'data': struct[module_type][module][param]
                    }
                for idx, prev in enumerate(prev_list):
                    cur_argument_idx = 'input{}'.format(idx)
                    step['arguments'][cur_argument_idx] = {
                        'data': prev
                    }
                pipeline['steps'].append(step)
            prev_list = new_prev_list
        pipeline['outputs'] = []
        for prev in prev_list:
            pipeline['outputs'].append({'data': prev})
        pipelines.append(pipeline)
    return pipelines


import networkx as nx
import numpy as np
from scipy.optimize import linear_sum_assignment
from ._similarity_flooding import similarity_flooding

#####
# Merges graphs
#
# Usage:
#
# from PipelineProfiler._graph_matching import pipeline_to_graph, merge_multiple_graphs
# from PipelineProfiler._plot_pipeline_node_link import plot_merged_pipeline
# pipelines = [pipeline_to_graph(x) for x in pipelines_problem["66_chlorineConcentration_problem"]]
# merged = merge_multiple_graphs(pipelines[:3])
# plot_merged_pipeline(merged)


def graph_data_accessor(node, graph_name):
    if 'hyperparams' in node:
        hyperparams = node['hyperparams']
    else:
        hyperparams = {}
    return {
        'python_path': node['primitive']['python_path'],
        'hyperparams': hyperparams,
        'graph_name': graph_name,
        'node_name': node['primitive']['name']
    }

def pipeline_to_graph(pipeline, name=''):
    if name == '':
        name = pipeline['pipeline_digest']
    G = nx.DiGraph(name=name)
    # Adding input node
    G.add_node('inputs.0', data=[{'python_path': 'Input', 'node_name': 'Input', 'hyperparams': {}, 'graph_name': name}])
    # Adding all steps
    steps = pipeline['steps']
    for idx, step in enumerate(steps):
        step_key = 'steps.' + str(idx)
        G.add_node(step_key, data = [graph_data_accessor(step, name)])
        for argument_key in step['arguments']:
            arguments = step['arguments'][argument_key]['data']
            if not isinstance(arguments, list):
                arguments = [arguments]
            for argument in arguments:
                argument_key = '.'.join(argument.split(".")[:2])
                G.add_edge(argument_key, step_key)
    # Adding output node
    G.add_node('outputs.0', data=[{'python_path': 'Output', 'node_name': 'Output', 'hyperparams': {}, 'graph_name': name}])
    out_node_input = '.'.join(pipeline['outputs'][0]['data'].split(".")[:2])
    G.add_edge(out_node_input, 'outputs.0')
    return G

def python_path_similarity(python_path1, python_path2):
    if python_path1 == python_path2:
        return 1
    else:
        split1 = python_path1.split(".")
        split2 = python_path2.split(".")
        if len(split1) > 2 and len(split2) > 2:
            if split1[2] == split2[2]:
                return 0.5
            else:
                return 0
        else:
            return 0

def compute_node_similarity_matrix(g1, g2):
    similarity = np.zeros([len(g1.nodes), len(g2.nodes)])
    for idx_g1, n_g1 in enumerate(g1.nodes):
        for idx_g2, n_g2 in enumerate(g2.nodes):
            datas1 = g1.nodes[n_g1]['data']
            datas2 = g2.nodes[n_g2]['data']
            similarities = []
            for data1 in datas1: # Node can have multiple primitives. Computing the max average similarity
                for data2 in datas2:
                    python_path1 = data1['python_path']
                    python_path2 = data2['python_path']
                    path_similarity = python_path_similarity(python_path1, python_path2)
                    # small hack to consider "neighborhood" (TODO: replace similarity flooding)
                    n_g1_proc = '.'.join(n_g1.split(".")[-2:])
                    n_g2_proc = '.'.join(n_g2.split(".")[-2:])
                    if n_g1_proc == n_g2_proc:
                        path_similarity += 0.05
                    similarities.append(path_similarity)
            similarity[idx_g1, idx_g2] = np.mean(similarities)
    return similarity

def compute_edit_cost_matrix(similarity_matrix, add_cost, del_cost):
    INF = 1000
    # compute
    n_nodes_g1 = similarity_matrix.shape[0]
    n_nodes_g2 = similarity_matrix.shape[1]
    len_g1_plus_g2 = n_nodes_g1 + n_nodes_g2
    cost_matrix = np.zeros([len_g1_plus_g2, len_g1_plus_g2])
    # Substitution cost (top left) of transforming g2 into g1
    cost_matrix[:n_nodes_g1, :n_nodes_g2] = np.max(similarity_matrix) - similarity_matrix # cost = 1 - similarity
    # Addition cost (top right) of adding nodes from g1
    addition = INF * np.ones([n_nodes_g1, n_nodes_g1])
    np.fill_diagonal(addition, add_cost)
    cost_matrix[:n_nodes_g1, n_nodes_g2:] = addition
    # Deletion cost (bottom left) of removing nodes from g2
    deletion = INF * np.ones([n_nodes_g2, n_nodes_g2])
    np.fill_diagonal(deletion, del_cost)
    cost_matrix[n_nodes_g1:, :n_nodes_g2] = deletion
    return cost_matrix


def compute_node_equivalence(g1, g2):
    helper_graph = nx.union(g1, g2, rename=('g1-','g2-')) #nodes start with `rename` prefix. Used to check for cycles
    equivalence_g1 = {} # maps nodes from g1 to g2
    equivalence_g2 = {} # maps nodes from g2 to g1
    len_g1 = len(g1.nodes)
    len_g2 = len(g2.nodes)
    similarity_matrix = compute_node_similarity_matrix(g1, g2)
    try:
        similarity_matrix = similarity_flooding(similarity_matrix, g1, g2, alpha = 0.1, n_iter = 50)
    except Exception:
        print("Similarity flooding failed. PCG graph has no nodes or edges.")
    edit_cost_matrix = compute_edit_cost_matrix(similarity_matrix, 0.4, 0.4)
    rows, cols = linear_sum_assignment(edit_cost_matrix)
    nodes_g1 = list(g1.nodes)
    nodes_g2 = list(g2.nodes)
    for pairs in zip(rows, cols):
        if pairs[0] < len_g1 and pairs[1] < len_g2:
            # nodes are equivalent
            eq_g1 = nodes_g1[pairs[0]]
            eq_g2 = nodes_g2[pairs[1]]
            merged_helper = nx.algorithms.minors.contracted_nodes(helper_graph, 'g1-'+eq_g1, 'g2-'+eq_g2, self_loops=False)
            try:
                if 'g1-inputs.0' in merged_helper:
                    cycles = nx.algorithms.cycles.find_cycle(merged_helper, source = 'g1-inputs.0')
                if 'g2-inputs.0' in merged_helper:
                    cycles = nx.algorithms.cycles.find_cycle(merged_helper, source = 'g2-inputs.0')
            except nx.NetworkXNoCycle as n:
                equivalence_g2[eq_g2] = eq_g1
                equivalence_g1[eq_g1] = eq_g2
                helper_graph = merged_helper
    return equivalence_g1, equivalence_g2

def dict_append(dictionary, key, value):
    if key in dictionary:
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]

def merge_graphs(g1, g2):
    equivalence_g1, equivalence_g2 = compute_node_equivalence(g1, g2)
    G = nx.DiGraph()
    # Adding g1 node data
    for node in g1.nodes:
        data = g1.nodes[node]["data"]
        if node not in equivalence_g1:
            node = "G1." + node
        G.add_node(node)
        G.nodes[node]["data"] = data

    # Adding g2 node data
    for node in g2.nodes:
        data = g2.nodes[node]["data"]
        if node not in equivalence_g2:
            node = "G2." + node
            G.add_node(node, data = data)
        else:
            node = equivalence_g2[node]
            G.nodes[node]["data"] = G.nodes[node]["data"] + data

    # Adding edges from g1
    for edge in g1.edges:
        source, dest = edge
        if source not in equivalence_g1:
            source = "G1." + source
        if dest not in equivalence_g1:
            dest = "G1." + dest
        G.add_edge(source, dest)

    # Merging edges from g2
    for edge in g2.edges:
        source, dest = edge
        if source not in equivalence_g2:
            source = "G2." + source
        else:
            source = equivalence_g2[source]
        if dest not in equivalence_g2:
            dest = "G2." + dest
        else:
            dest = equivalence_g2[dest]
        G.add_edge(source, dest)
    return G

def merge_multiple_graphs(graphs):
    merged = merge_graphs(graphs[0], graphs[1])
    for graph in graphs[2:]:
        merged = merge_graphs(merged, graph)
    return merged


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pipelineprofiler",
    version="0.1.16",
    author="Jorge Piazentin Ono, Sonia Castelo, Roque Lopez, Enrico Bertini, Juliana Freire, Claudio Silva",
    author_email="jorgehpo@nyu.edu",
    description="Pipeline Profiler tool. Enables the exploration of D3M pipelines in Jupyter Notebooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VIDA-NYU/PipelineVis",
    packages=find_packages(exclude=['js', 'node_modules']),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "python-dateutil",
        "numpy",
        "scipy",
        "scikit-learn",
        "networkx",
        "notebook"
    ]
)



import pkg_resources
import string
import numpy as np
import json
import networkx as nx

def id_generator(size=15):
    """Helper function to generate random div ids. This is useful for embedding
    HTML into ipython notebooks."""
    chars = list(string.ascii_uppercase)
    return ''.join(np.random.choice(chars, size, replace=True))


def make_html(data_dict, draw_function):
	lib_path = pkg_resources.resource_filename(__name__, "build/pipelineVis.js")
	bundle = open(lib_path).read()
	html_all = """
	<html>
	<head>
	</head>
	<body>
	    <script>
	    {bundle}
	    </script>
	    <div id="{id}">
	    </div>
	    <script>
	        pipelineVis.{draw_function}("#{id}", {data_dict});
	    </script>
	</body>
	</html>
	""".format(bundle=bundle, draw_function=draw_function, id=id_generator(), data_dict=json.dumps(data_dict))
	return html_all

def plot_pipeline_node_link(data_dict):
    """Takes as input an array of model expl objects (mldiff.compute_exp_obj) and the dataset, and plots coocurring rules
    in the IPython cell."""
    from IPython.core.display import display, HTML
    html_all = make_html(data_dict, draw_function="renderPipelineNodeLink")
    display(HTML(html_all))

def plot_merged_pipeline(merged):
    """Takes as input networkx.DiGraph and renders a graph in the IPython Cell"""
    from IPython.core.display import display, HTML
    data_dict = nx.readwrite.json_graph.node_link_data(merged)
    html_all = make_html(data_dict, draw_function="renderMergedPipeline")
    display(HTML(html_all))


import pkg_resources
import json

def get_heartstatlog_data():
    data_path = pkg_resources.resource_filename(__name__, "data/teaser_data_heartstatlog.json")
    with open(data_path, "r") as f:
        data = json.load(f)
    return data